; File:		avrispv2\main.asm
; Device:	AVR
; Created:	2024-06-20 01:12:09
; Version:	2025-01-28
; Author:	Johannes Fechner

; Devices to be supported (in future):
.equ T3227 = 0
.equ A64DD32 = 1 ; not supported yet
.equ M88A = 2 ; not supported yet

; Device selection:
.equ MCU = T3227

.nolist
.if MCU == T3227
.include "C:\Program Files\Microchip\MPLABX\v6.20\packs\Microchip\ATtiny_DFP\3.1.260\avrasm\inc\tn3227def.inc"
.else
.error "Device not supported yet."
.endif
.include "stk500v2-isp.inc"
.include "..\macros.inc"
.list

.equ DEBUG = 1 ; Whether to send debug messages via UART1 (not available on some devices)
.equ SPI_INTERBYTE_DELAY_MS = 1 ; Count of milliseconds to wait after each SPI transfer

; == Hardware definitions ==
.if MCU == A64DD32
; CPU clock frequency:
.equ F_CPU = 20_000_000 ; Hz

; Target MCU Reset:
.equ RESET_VPORT_DIR = VPORTD_DIR
.equ RESET_VPORT_OUT = VPORTD_OUT
.equ RESET_bp = 1

; UART for STK500v2 protocol:
.equ UART_STK_BPS = 115_200 ; bits/s
.equ UART_STK_VPORT_DIR = VPORTA_DIR
.equ UART_STK_TXD_bp = 0

.if DEBUG
; UART for sending debugging messages:
.equ UART_DBG_BPS = 115_200 ; bits/s
.equ UART_DBG_VPORT_DIR = VPORTC_DIR
.equ UART_DBG_TXD_bp = 0
.endif ; DEBUG

; SPI for communication with target MCU:
.equ SPI_VPORT_DIR = VPORTA_DIR
.equ SPI_MOSI_bp = 4
.equ SPI_SCK_bp = 6

.elif MCU == T3227
; CPU clock frequency:
.equ F_CPU = 20_000_000 ; Hz

; Target MCU Reset:
.equ RESET_VPORT_DIR = VPORTA_DIR
.equ RESET_VPORT_OUT = VPORTA_OUT
.equ RESET_bp = 6

; UART for STK500v2 protocol:
.equ UART_STK_BPS = 115_200 ; bits/s
.equ UART_STK_VPORT_DIR = VPORTB_DIR
.equ UART_STK_TXD_bp = 2

.if DEBUG
; UART for sending debugging messages:
.equ UART_DBG_BPS = 115_200 ; bits/s
; Alternate pin position:
.equ UART_DBG_VPORT_DIR = VPORTC_DIR
.equ UART_DBG_TXD_bp = 2
.endif ; DEBUG

; SPI for communication with target MCU:
.equ SPI_VPORT_DIR = VPORTA_DIR
.equ SPI_MOSI_bp = 1
.equ SPI_SCK_bp = 3

.elif MCU == M88A
; CPU clock frequency:
.equ F_CPU = 18_432_000 ; Hz

; Target MCU Reset:
.equ RESET_VPORT_DIR = DDRB
.equ RESET_VPORT_OUT = PORTB
.equ RESET_bp = 1

; UART for communication with host PC:
.equ UART_STK_BPS = 115_200 ; bits/s
.equ UART_STK_VPORT_DIR = DDRD
.equ UART_STK_TXD_bp = 1

.if DEBUG
.error "DEBUG not available for chosen device."
.endif

; SPI for communication with target MCU:
.equ SPI_VPORT_DIR = DDRB
.equ SPI_MOSI_bp = 3
.equ SPI_SCK_bp = 5

.else
.error "Device not currently supported."
.endif

; == STK500v2 protocol definitions ==
.equ STK500V2_MAX_MSG_SIZE = 275

; == Register name definitions ==
.def zero = r2		; Contains always 0x00.
.def wr0 = r3
.def wr1 = r4
.def wr2 = r5
.def wr3 = r6
.def wr4 = r14
.def wr5 = r15
.def wr6 = r0
.def wr7 = r1
.def wri0 = r16
.def wri1 = r17
.def wri2 = r18
.def wri3 = r19
.def wri4 = R20
.def event = r21
.def countdown = r23
.def wriL = r24
.def wriH = r25

; == Values of register 'event' ==
.equ EVENT_RX_SUCCESS = 1 ; A message was successfully received.
.equ EVENT_CKSUM_ERROR = 2 ; A message with invalid checksum was received.

; == Data memory usage ==
.dseg
.org SRAM_START
stk500v2State_sram: .byte 1
sequenceNumber_sram: .byte 1
messageSize_sram: .byte 2
messageBody_sram: .byte STK500V2_MAX_MSG_SIZE
messageBodyCntr_sram: .byte 2
checksum_sram: .byte 1
controllerInit_sram: .byte 1
address_sram: .byte 4 ; LSB at lowest index ('normal' order)
pageIndex_sram: .byte 1

.cseg
.org 0
	rjmp	reset

; == Interrupt vectors ==
.org TCA0_OVF_vect
	rjmp	tcaOvfIsr
.org USART0_RXC_vect
	rjmp	uartRxcIsr

.org INT_VECTORS_SIZE

; == Subroutines ==
.include "routines.asm"

; == Constant data ==
signature_flash: .db STATUS_CMD_OK, 8, "AVRISP_2"
.equ SIGNATURE_LENGTH = 10

.if DEBUG
checksumErr_flash:			.db "Checksum Error", '\r', '\n', 0, 0
receivedCmd_flash:			.db "Received command ", 0
signOn_flash:				.db "SIGN_ON", '\r', '\n', 0
setParam_flash:				.db "SET_PARAMETER ", 0, 0
lineBreak_flash:			.db '\r', '\n', 0, 0
getParam_flash:				.db "GET_PARAMETER ", 0, 0
loadAddr_flash:				.db "LOAD_ADDRESS ", 0
enterProgmode_flash:		.db "ENTER_PROGMODE", '\r', '\n', 0, 0
leaveProgmode_flash:		.db "LEAVE_PROGMODE", '\r', '\n', 0, 0
chipErase_flash:			.db "CHIP_ERASE", '\r', '\n', 0, 0
pgmFlashEeprom_flash:		.db "PROGRAM_FLASH or PROGRAM_EEPROM", '\r', '\n', 0
readFlashEeprom_flash:		.db "READ_FLASH or READ_EEPROM", '\r', '\n', 0
pgmFuse_flash:				.db "PROGRAM_FUSE", '\r', '\n', 0, 0
readFuseSignature_flash:	.db "READ_FUSE or READ_SIGNATURE", '\r', '\n', 0
unknown_flash:				.db "(unknown)", '\r', '\n', 0
.endif ; DEBUG

; == Execution entry point after reset ==
reset:

; == Initialize registers ==
	clr		zero
	clr		event

; == Disable clock prescaler ==
	ldi		wri0, CPU_CCP_IOREG_gc
	xout	CPU_CCP, wri0
	xin		wri0, CLKCTRL_MCLKCTRLB
	andi	wri0, ~CLKCTRL_PEN_bm
	xout	CLKCTRL_MCLKCTRLB, wri0
	
; == Initialize UART ==
; Connected with host PC.
.equ BAUD0 = ROUND(4.0*F_CPU/(1.0*UART_STK_BPS))
	ldi		wri0, low(BAUD0)
	xout	USART0_BAUDL, wri0
	ldi		wri0, high(BAUD0)
	xout	USART0_BAUDH, wri0
	ldi		wri0, USART_RXCIE_bm
	xout	USART0_CTRLA, wri0
	ldi		wri0, USART_RXEN_bm | USART_TXEN_bm
	xout	USART0_CTRLB, wri0
	sbi		UART_STK_VPORT_DIR, UART_STK_TXD_bp
	
.if DEBUG
.if MCU == T3227
.message "Selecting alternate pin positions for USART1."
	ldi		wri0, PORTMUX_USART1_ALT1_gc
	xout	PORTMUX_USARTROUTEA, wri0
.endif ; MCU

; == Initialize Debugging UART ==
.equ BAUD1 = ROUND(4.0*F_CPU/(1.0*UART_DBG_BPS))
	ldi		wri0, low(BAUD1)
	xout	USART1_BAUDL, wri0
	ldi		wri0, high(BAUD1)
	xout	USART1_BAUDH, wri0
	ldi		wri0, USART_TXEN_bm
	xout	USART1_CTRLB, wri0
	sbi		UART_DBG_VPORT_DIR, UART_DBG_TXD_bp
.endif ; DEBUG

; == Initialize SPI ==
; Connected with target/client MCU.
	ldi		wri0, SPI_SSD_bm
	xout	SPI0_CTRLB, wri0
	ldi		wri0, SPI_MASTER_bm | SPI_PRESC_1_bm | SPI_PRESC_0_bm | SPI_ENABLE_bm
	xout	SPI0_CTRLA, wri0
	
; == Initialize TCA0 ==
; 1-ms periodic interrupt.
.equ TCA0_PERIOD = ROUND(F_CPU*1.0e-3)
	ldi		wri0, low(TCA0_PERIOD)
	xout	TCA0_SINGLE_PERL, wri0
	ldi		wri0, high(TCA0_PERIOD)
	xout	TCA0_SINGLE_PERH, wri0
	ldi		wri0, TCA_SINGLE_OVF_bm
	xout	TCA0_SINGLE_INTCTRL, wri0

; == Initialize SRAM variables ==
	sts		stk500v2State_sram, zero
	sts		controllerInit_sram, zero
	sts		pageIndex_sram, zero

; == Enable interrupts ==
	sei
	
loop:
	cpi		event, EVENT_RX_SUCCESS
	breq	validMsg
	cpi		event, EVENT_CKSUM_ERROR
	brne	loop
	
; A message with wrong checksum was received.
	clr		event
.if DEBUG
	ldiz	2*checksumErr_flash
	rcall	sendDbgMsg
.endif ; DEBUG
	sts		messageBodyCntr_sram, zero
	ldi		wri0, ANSWER_CKSUM_ERROR
	rcall	appendByteToMsgBody
	ldi		wri0, STATUS_CKSUM_ERROR
	rcall	appendByteToMsgBody
	rcall	returnAnswer
	rjmp	loop
	
validMsg:
; A valid message was received.
	clr		event
.if DEBUG
	ldiz	2*receivedCmd_flash
	rcall	sendDbgMsg
.endif ; DEBUG
	ldi		wri0, 1 ; Echo Command ID as first byte.
	sts		messageBodyCntr_sram, wri0
; Branch according to command ID:
	lds		wri0, messageBody_sram ; Load Command_ID, always first byte of message.
	cpi		wri0, CMD_SIGN_ON
	brne	cmdSetParameter
.if DEBUG
	ldiz	2*signOn_flash
	rcall	sendDbgMsg
.endif ; DEBUG
	ldi		wri0, SIGNATURE_LENGTH
	ldiz	2*signature_flash
	rcall	appendFlashBytesToMsgBody
	rcall	returnAnswer
	rjmp	loop
	
cmdSetParameter:
	cpi		wri0, CMD_SET_PARAMETER
	brne	cmdGetParameter
.if DEBUG
	ldiz	2*setParam_flash
	rcall	sendDbgMsg
.endif ; DEBUG
	lds		wri0, messageBody_sram + CMD_SET_PARAMETER_PARAMID
.if DEBUG
	rcall	sendDbgHex
	ldiz	2*lineBreak_flash
	rcall	sendDbgMsg
.endif ; DEBUG
	cpi		wri0, PARAM_CONTROLLER_INIT
	brne	cmdSetParameter0
	lds		wri0, messageBody_sram + CMD_SET_PARAMETER_VALUE
	sts		controllerInit_sram, wri0
cmdSetParameter0:
; Dummy catch-all which does nothing but responding ACK:
	ldi		wri0, STATUS_CMD_OK
	rcall	appendByteToMsgBody
	rcall	returnAnswer
	rjmp	loop

cmdGetParameter:
	cpi		wri0, CMD_GET_PARAMETER
	brne	cmdLoadAddress
.if DEBUG
	ldiz	2*getParam_flash
	rcall	sendDbgMsg
.endif ; DEBUG
	lds		wri0, messageBody_sram + CMD_GET_PARAMETER_PARAMID
.if DEBUG
	rcall	sendDbgHex
	ldiz	2*lineBreak_flash
	rcall	sendDbgMsg
.endif ; DEBUG
	cpi		wri0, PARAM_BUILD_NUMBER_LOW
	brne	cmdGetParameter0
	ldi		wri0, STATUS_CMD_OK
	rcall	appendByteToMsgBody
	ldi		wri0, 0xCD ; Dummy number.
	rcall	appendByteToMsgBody
	rcall	returnAnswer
	rjmp	loop
cmdGetParameter0:
	cpi		wri0, PARAM_BUILD_NUMBER_HIGH
	brne	cmdGetParameter1
	ldi		wri0, STATUS_CMD_OK
	rcall	appendByteToMsgBody
	ldi		wri0, 0xAB ; Dummy number.
	rcall	appendByteToMsgBody
	rcall	returnAnswer
	rjmp	loop
cmdGetParameter1:
	cpi		wri0, PARAM_HW_VER
	brne	cmdGetParameter2
	ldi		wri0, STATUS_CMD_OK
	rcall	appendByteToMsgBody
	ldi		wri0, 0
	rcall	appendByteToMsgBody
	rcall	returnAnswer
	rjmp	loop
cmdGetParameter2:
	cpi		wri0, PARAM_SW_MAJOR
	brne	cmdGetParameter3
	ldi		wri0, STATUS_CMD_OK
	rcall	appendByteToMsgBody
	ldi		wri0, 0
	rcall	appendByteToMsgBody
	rcall	returnAnswer
	rjmp	loop
cmdGetParameter3:
	cpi		wri0, PARAM_SW_MINOR
	brne	cmdGetParameter4
	ldi		wri0, STATUS_CMD_OK
	rcall	appendByteToMsgBody
	ldi		wri0, 0
	rcall	appendByteToMsgBody
	rcall	returnAnswer
	rjmp	loop
cmdGetParameter4:
	cpi		wri0, PARAM_CONTROLLER_INIT
	brne	cmdGetParameter5
	ldi		wri0, STATUS_CMD_OK
	rcall	appendByteToMsgBody
	lds		wri0, controllerInit_sram
	rcall	appendByteToMsgBody
	rcall	returnAnswer
	rjmp	loop
cmdGetParameter5:
	ldi		wri0, STATUS_CMD_FAILED
	rcall	appendByteToMsgBody
	rcall	returnAnswer
	rjmp	loop

; CMD_OSCCAL not implemented yet.

cmdLoadAddress:
	cpi		wri0, CMD_LOAD_ADDRESS
	brne	cmdEnterProgmode
.if DEBUG
	ldiz	2*loadAddr_flash
	rcall	sendDbgMsg
.endif ; DEBUG
; Copy the supplied 4-byte address:
	ldiz	messageBody_sram + CMD_LOAD_ADDRESS_ADDRMSB
	ldiy	address_sram + 4
	clr		wri1
cmdLoadAddress0:
	ld		wri0, Z+
	st		-Y, wri0
.if DEBUG
	rcall	sendDbgHex
.endif ; DEBUG
	inc		wri1
	cpi		wri1, 4
	brlo	cmdLoadAddress0
.if DEBUG
	ldiz	2*lineBreak_flash
	rcall	sendDbgMsg
.endif ; DEBUG
; Check whether a Load Extended Address is requested:
	lds		wri0, address_sram+3 ; Load address MSB.
	andi	wri0, 0x80
	breq	cmdLoadAddress1
	ldi		wri0, 0x4D ; ISP instruction byte #1
	rcall	spiTxRx
	ldi		wri0, 0x00 ; ISP instruction byte #2
	rcall	spiTxRx
	lds		wri0, address_sram+2 ; ISP instruction byte #3 == extended address
	rcall	spiTxRx
	ldi		wri0, 0x00 ; ISP instruction byte #4
	rcall	spiTxRx
cmdLoadAddress1:
; Return answer:
	ldi		wri0, STATUS_CMD_OK
	rcall	appendByteToMsgBody
	rcall	returnAnswer
	rjmp	loop
	
; CMD_FIRMWARE_UPGRADE won't be implemented as it is Atmel-specific.

cmdEnterProgmode:
	cpi		wri0, CMD_ENTER_PROGMODE_ISP
	brne	cmdLeaveProgmode
.if DEBUG
	ldiz	2*enterProgmode_flash
	rcall	sendDbgMsg
.endif ; DEBUG
	sbi		RESET_VPORT_DIR, RESET_bp ; pull RESET low
	ldi		countdown, 1
	rcall	waitMs
	sbi		SPI_VPORT_DIR, SPI_SCK_bp
	sbi		SPI_VPORT_DIR, SPI_MOSI_bp
	ldi		countdown, 100
	rcall	waitMs
; Apply a positive pulse of at least two target CPU clock cycles:
; (assuming F_CPU_TARGET >= 3.6864 MHz)
	cbi		RESET_VPORT_DIR, RESET_bp
	wait_ns_ 600, 0
	sbi		RESET_VPORT_DIR, RESET_bp
	ldi		countdown, 100
	rcall	waitMs
	ldiz	messageBody_sram + CMD_ENTER_PROGMODE_ISP_CMD1
	clr		wri1 ; Loop counter.
	lds		wri2, messageBody_sram + CMD_ENTER_PROGMODE_ISP_POLLVALUE
	lds		wri3, messageBody_sram + CMD_ENTER_PROGMODE_ISP_POLLINDEX
	dec		wri3 ; Adjust to zero-based index.
	clr		wri4 ; To be changed if pollValue doesn't match.
cmdEnterProgmode0:
	ld		wri0, Z+
	rcall	spiTxRx
	cp		wri3, wri1
	brne	cmdEnterProgmode1
	cp		wri2, wri0
	breq	cmdEnterProgmode1
	inc		wri4
cmdEnterProgmode1:
	inc		wri1
	cpi		wri1, 4
	brlo	cmdEnterProgmode0
	tst		wri4
	brne	cmdEnterProgmode2
	ldi		wri0, STATUS_CMD_OK
	rcall	appendByteToMsgBody
	rcall	returnAnswer
	rjmp	loop
cmdEnterProgmode2:
	ldi		wri0, STATUS_CMD_FAILED
	rcall	appendByteToMsgBody
	rcall	returnAnswer
	rjmp	loop
	
cmdLeaveProgmode:
	cpi		wri0, CMD_LEAVE_PROGMODE_ISP
	brne	cmdChipErase
.if DEBUG
	ldiz	2*leaveProgmode_flash
	rcall	sendDbgMsg
.endif ; DEBUG
	cbi		SPI_VPORT_DIR, SPI_MOSI_bp
	cbi		SPI_VPORT_DIR, SPI_SCK_bp
	ldi		countdown, 1
	rcall	waitMs
	cbi		RESET_VPORT_DIR, RESET_bp
	ldi		wri0, STATUS_CMD_OK
	rcall	appendByteToMsgBody
	rcall	returnAnswer
	rjmp	loop

cmdChipErase:
	cpi		wri0, CMD_CHIP_ERASE_ISP
	brne	cmdProgramFlash
.if DEBUG
	ldiz	2*chipErase_flash
	rcall	sendDbgMsg
.endif ; DEBUG
; Send the four ISP command bytes:
	ldiz	messageBody_sram + CMD_CHIP_ERASE_ISP_CMD1
	clr		wri1 ; Loop counter.
cmdChipErase0:
	ld		wri0, Z+
	rcall	spiTxRx
	inc		wri1
	cpi		wri1, 4
	brlo	cmdChipErase0
; Wait for target device to get ready:
	lds		wri1, messageBody_sram + CMD_CHIP_ERASE_ISP_POLLMETHOD
	tst		wri1
	brne	cmdChipErase1 ; Branch if polling shall be used.
	lds		countdown, messageBody_sram+1 ; Load the supplied delay time (in ms).
	rcall	waitMs
	rjmp	cmdChipErase2
cmdChipErase1:
	rcall	pollRdyBsy
cmdChipErase2:
	ldi		wri0, STATUS_CMD_OK
	rcall	appendByteToMsgBody
	rcall	returnAnswer
	rjmp	loop
	
cmdProgramFlash:
	cpi		wri0, CMD_PROGRAM_FLASH_ISP
	breq	cmdProgramFlash7
	cpi		wri0, CMD_PROGRAM_EEPROM_ISP
	breq	cmdProgramFlash7
	rjmp	cmdReadFlash
cmdProgramFlash0:
	ldi		wri0, STATUS_CMD_FAILED
	rcall	appendByteToMsgBody
	rcall	returnAnswer
	rjmp	loop
cmdProgramFlash7:
.if DEBUG
	ldiz	2*pgmFlashEeprom_flash
	rcall	sendDbgMsg
.endif ; DEBUG
	mov		wri3, wri0 ; Make a copy of the STK command.
; Load NumBytes and make sure it's not too big:
	lds		wriH, messageBody_sram + CMD_PROGRAM_FLASH_ISP_NUMBYTESH
	lds		wriL, messageBody_sram + CMD_PROGRAM_FLASH_ISP_NUMBYTESL
	movw	wri0, wriL
.equ PROGRAM_FLASH_NUMBYTES_TSTVAL = STK500V2_MAX_MSG_SIZE-9 ; BufSize-OtherMsgBytes+1
	subi	wri0, low(PROGRAM_FLASH_NUMBYTES_TSTVAL)
	sbci	wri1, high(PROGRAM_FLASH_NUMBYTES_TSTVAL)
	brsh	cmdProgramFlash0 ; Branch if NumBytes is greater than buffer size allows.
; NumBytes is OK.
	lds		wri1, messageBody_sram + CMD_PROGRAM_FLASH_ISP_MODE
	bst		wri1, CMD_PROGRAM_FLASH_ISP_MODE_WP_bp ; Word/Page Mode (0 = word, 1 = page)
	brts	cmdProgramFlash11
	lds		wr0, address_sram
	lds		wr1, address_sram+1
	rjmp	cmdProgramFlash12
cmdProgramFlash11:
	lds		wr0, pageIndex_sram
	clr		wr1
cmdProgramFlash12:
; In case of flash programming, convert to byte address wr2:wr1:wr0
; and prepare H bit mask:
	clr		wr2
	cpi		wri3, CMD_PROGRAM_FLASH_ISP
	brne	cmdProgramFlash14
	lsl		wr0
	rol		wr1
	rol		wr2
	ldi		wri0, 1<<CMD_PROGRAM_FLASH_ISP_CMD1_H_bp
	mov		wr3, wri0
cmdProgramFlash14:
	ldiz	messageBody_sram + CMD_PROGRAM_FLASH_ISP_DATA
cmdProgramFlash2:
	lds		wri0, messageBody_sram + CMD_PROGRAM_FLASH_ISP_CMD1
	mov		wr4, wr0
	mov		wr5, wr1
	mov		wr6, wr2
; In case of flash programming, convert back to word address
; and handle H bit:
	cpi		wri3, CMD_PROGRAM_FLASH_ISP
	brne	cmdProgramFlash13
	lsr		wr6
	ror		wr5 ; MSB
	ror		wr4 ; LSB
	sbrc	wr0, 0 ; skip setting H bit if address is even
	or		wri0, wr3
cmdProgramFlash13:
	rcall	spiTxRx
	mov		wri0, wr5 ; ISP instruction byte #2 == address MSB
	rcall	spiTxRx
	mov		wri0, wr4 ; ISP instruction byte #3 == address LSB
	rcall	spiTxRx
	ld		wri0, Z+ ; ISP instruction byte #4 == data
	rcall	spiTxRx
; Increment byte address:
	ldi		wri0, -1
	sub		wr0, wri0
	sbc		wr1, zero
	sbc		wr2, zero
; Modify loop counter:
	sbiw	wriL, 1
	brne	cmdProgramFlash2
; If in page mode, write back page index:
	brtc	cmdProgramFlash3 ; skip if in word mode
	cpi		wri3, CMD_PROGRAM_FLASH_ISP
	brne	cmdProgramFlash17
	lsr		wr2
	ror		wr1 ; MSB
	ror		wr0 ; LSB
cmdProgramFlash17:
	sts		pageIndex_sram, wr0
; If instructed so, issue the page write command:
	bst		wri1, 7 ; Whether to write page (0 = no, 1 = yes)
	brtc	cmdProgramFlash3 ; skip page write
	lds		wri0, messageBody_sram + CMD_PROGRAM_FLASH_ISP_CMD2
	rcall	spiTxRx
	lds		wri0, address_sram+1 ; ISP instruction byte #2 == address MSB
	rcall	spiTxRx
	lds		wri0, address_sram ; ISP instruction byte #3 == address LSB
	rcall	spiTxRx
	ldi		wri0, 0x00 ; ISP instruction byte #4
	rcall	spiTxRx
	sts		pageIndex_sram, zero ; reset page index
cmdProgramFlash3:
; Wait for the target device to process the write command:
	bst		wri1, 4 ; Whether to use time delay
	brtc	cmdProgramFlash4
	lds		countdown, messageBody_sram + CMD_PROGRAM_FLASH_ISP_DELAY
	rcall	waitMs
	rjmp	cmdProgramFlash6
cmdProgramFlash4:
	bst		wri1, 5 ; Whether to use value polling
	brtc	cmdProgramFlash5
; To be implemented later, just use constant delay for now:
	ldi		countdown, 5
	rcall	waitMs
	rjmp	cmdProgramFlash6
cmdProgramFlash5:
	bst		wri1, 6 ; Whether to use RDY/BSY polling
	brtc	cmdProgramFlash6
	rcall	pollRdyBsy
cmdProgramFlash6:
; Send answer:
	ldi		wri0, STATUS_CMD_OK
	rcall	appendByteToMsgBody
	rcall	returnAnswer
	rjmp	loop
	
cmdReadFlash:
	cpi		wri0, CMD_READ_FLASH_ISP
	breq	cmdReadFlash1
	cpi		wri0, CMD_READ_EEPROM_ISP
	breq	cmdReadFlash1
	rjmp	cmdProgramFuse
cmdReadFlash1:
.if DEBUG
	ldiz	2*readFlashEeprom_flash
	rcall	sendDbgMsg
.endif ; DEBUG
	mov		wri4, wri0 ; Make a copy of the STK command.
; Load NumBytes and make sure it's not too big:
	lds		wriH, messageBody_sram + CMD_READ_FLASH_ISP_NUMBYTESH
	lds		wriL, messageBody_sram + CMD_READ_FLASH_ISP_NUMBYTESL
	movw	wri0, wriL
.equ READ_FLASH_NUMBYTES_TSTVAL = STK500V2_MAX_MSG_SIZE-2 ; AnswerID, Status1, and Status2 must fit into buffer, too.
	subi	wri0, low(READ_FLASH_NUMBYTES_TSTVAL)
	sbci	wri1, high(READ_FLASH_NUMBYTES_TSTVAL)
	brsh	cmdReadFlash0 ; Branch if NumBytes is greater than buffer size allows.
; NumBytes is OK.
	lds		wri1, messageBody_sram + CMD_READ_FLASH_ISP_CMD1
	lds		wr0, address_sram ; address LSB
	lds		wr1, address_sram+1 ; address MSB
	ldi		wri0, STATUS_CMD_OK
	rcall	appendByteToMsgBody
; In case of flash programming, convert to byte address wr2:wr1:wr0
; and prepare H bit mask:
	clr		wr2
	cpi		wri4, CMD_PROGRAM_FLASH_ISP
	brne	cmdReadFlash2
	lsl		wr0
	rol		wr1
	rol		wr2
	ldi		wri0, 1<<CMD_PROGRAM_FLASH_ISP_CMD1_H_bp
	mov		wr3, wri0
cmdReadFlash2:
	mov		wri0, wri1
	mov		wr4, wr0
	mov		wr5, wr1
	mov		wr6, wr2
; In case of flash programming, convert back to word address
; and handle H bit:
	cpi		wri3, CMD_PROGRAM_FLASH_ISP
	brne	cmdReadFlash5
	lsr		wr6
	ror		wr5 ; MSB
	ror		wr4 ; LSB
	sbrc	wr0, 0 ; skip setting H bit if address is even
	or		wri0, wr3
cmdReadFlash5:
	rcall	spiTxRx
	mov		wri0, wr5
	rcall	spiTxRx
	mov		wri0, wr4
	rcall	spiTxRx
	ldi		wri0, 0x00
	rcall	spiTxRx
	rcall	appendByteToMsgBody
; Increment byte address:
	ldi		wri0, -1
	sub		wr0, wri0
	sbc		wr1, zero
	sbc		wr2, zero
; Modify loop counter:
	sbiw	wriL, 1
	brne	cmdReadFlash2
; Send answer:
	ldi		wri0, STATUS_CMD_OK
	rcall	appendByteToMsgBody
	rcall	returnAnswer
	rjmp	loop
cmdReadFlash0:
	ldi		wri0, STATUS_CMD_FAILED
	rcall	appendByteToMsgBody
	rcall	returnAnswer
	rjmp	loop

cmdProgramFuse:
	cpi		wri0, CMD_PROGRAM_FUSE_ISP
	brne	cmdReadFuse
.if DEBUG
	ldiz	2*pgmFuse_flash
	rcall	sendDbgMsg
.endif ; DEBUG
	ldiz	messageBody_sram + CMD_PROGRAM_FUSE_ISP_CMD1
	clr		wri1 ; Loop counter.
cmdProgramFuse0:
	ld		wri0, Z+
	rcall	spiTxRx
	inc		wri1
	cpi		wri1, 4
	brlo	cmdProgramFuse0
; A delay could be inserted here to make sure the target device
; can process the write command before the next one.
	ldi		wri0, STATUS_CMD_OK
	rcall	appendByteToMsgBody
	rcall	appendByteToMsgBody
	rcall	returnAnswer
	rjmp	loop

cmdReadFuse:
	cpi		wri0, CMD_READ_FUSE_ISP
	breq	cmdReadFuse2
	cpi		wri0, CMD_READ_SIGNATURE_ISP
	brne	cmdUnknown
cmdReadFuse2:
.if DEBUG
	ldiz	2*readFuseSignature_flash
	rcall	sendDbgMsg
.endif ; DEBUG
	ldiz	messageBody_sram + CMD_READ_FUSE_CMD1
	clr		wri1 ; Loop counter.
	lds		wri3, messageBody_sram + CMD_READ_FUSE_RETADDR
	dec		wri3 ; Adjust to zero-based index.
cmdReadFuse0:
	ld		wri0, Z+
	rcall	spiTxRx
	cp		wri3, wri1
	brne	cmdReadFuse1
	mov		wri2, wri0 ; Copy the returned byte.
cmdReadFuse1:
	inc		wri1
	cpi		wri1, 4
	brlo	cmdReadFuse0
	ldi		wri0, STATUS_CMD_OK
	rcall	appendByteToMsgBody
	mov		wri0, wri2
	rcall	appendByteToMsgBody
	ldi		wri0, STATUS_CMD_OK
	rcall	appendByteToMsgBody
	rcall	returnAnswer
	rjmp	loop

cmdUnknown:
; Command either invalid or not supported (yet).
.if DEBUG
	ldiz	2*unknown_flash
	rcall	sendDbgMsg
.endif ; DEBUG
	ldi		wri0, STATUS_CMD_UNKNOWN
	rcall	appendByteToMsgBody
	rcall	returnAnswer
	rjmp	loop
