; File:		avrispv2\routines.asm
; Device:	AVR
; Created:	2024-06-20 01:12:09
; Version:	2025-02-01
; Author:	Johannes Fechner

; == Routine: Wait for previous USART0 transmission to finish and send next byte ==
; Parameter: wri0, the byte to be sent.
sendStkByte:
	push	wri0
; Make sure the transmit data registers are empty:
sendStkByte0:
	xin		wri0, USART0_STATUS
	sbrs	wri0, USART_DREIF_bp
	rjmp	sendStkByte0
	pop		wri0
	xout	USART0_TXDATAL, wri0
	ret

.if DEBUG
; == Routine: Wait for previous USART1 transmission to finish and send next byte ==
; Parameter: wri0, the byte to be sent.
sendDbgByte:
	push	wri0
; Make sure the transmit data registers are empty:
sendDbgByte0:
	xin		wri0, USART1_STATUS
	sbrs	wri0, USART_DREIF_bp
	rjmp	sendDbgByte0
	pop		wri0
	xout	USART1_TXDATAL, wri0
	ret

; === Routine: Send string via USART1 ===
; Parameter: Z pointer, start address of the null-terminated string.
; String length is limited to 255 characters.
sendDbgMsg:
	push	wri0
	push	wri1
	clr		wri1
sendDbgMsg_loop:
	lpm		wri0, Z+
	tst		wri0
	breq	sendDbgMsg_end
	rcall	sendDbgByte
	inc		wri1
	cpi		wri1, $FF
	breq	sendDbgMsg_end
	rjmp	sendDbgMsg_loop
sendDbgMsg_end:
	pop		wri1
	pop		wri0
	ret

; === Routine: Send byte in hexadecimal via USART1 ===
; Send 8-bit unsigned number as two hex digits.
; Parameter: wri0.
sendDbgHex:
	push	wri0
	swap	wri0
	andi	wri0, 0x0F
	rcall	sendDbgHexDigit
	pop		wri0
	push	wri0
	andi	wri0, 0x0F
	rcall	sendDbgHexDigit
	pop		wri0
	ret
sendDbgHexDigit:
	cpi		wri0, 10
	brlt	sendDbgHexDigit0
	subi	wri0, -('A'-'9'-1)
sendDbgHexDigit0:
	subi	wri0, -'0'
	rcall	sendDbgByte
	ret
.endif ; DEBUG

; == Routine: Transmit byte via SPI and wait for finishing ==
; Parameter: wri0, the byte to be transmitted.
; Return: wri0, the byte read from SPI.
spiTxRx:
	xout	SPI0_DATA, wri0
spiTxRx0:
	xin		wri0, SPI0_INTFLAGS
	andi	wri0, SPI_IF_bm
	breq	spiTxRx0
	xin		wri0, SPI0_DATA
.if SPI_INTERBYTE_DELAY_MS
	ldi		countdown, SPI_INTERBYTE_DELAY_MS
	rcall	waitMs
.endif ; SPI_INTERBYTE_DELAY_MS
	ret

; == Routine: Wait ms ==
; Parameter: countdown, the count of ms to wait; will contain zero afterwards.
waitMs:
	push	wri0
	xin		wri0, TCA0_SINGLE_CTRLA
	ori		wri0, TCA_SINGLE_ENABLE_bm
	xout	TCA0_SINGLE_CTRLA, wri0
waitMs0:
	tst		countdown
	brne	waitMs0
	xin		wri0, TCA0_SINGLE_CTRLA
	andi	wri0, ~TCA_SINGLE_ENABLE_bm
	xout	TCA0_SINGLE_CTRLA, wri0
	pop		wri0
	ret

; == Routine: Append byte to message body ==
; Parameter: wri0, the byte to be appended.
appendByteToMsgBody:
	push	wriL
	push	wriH
	push	ZL
	push	ZH
; Check whether messageBodyCntr_sram is in allowed range:
	lds		wriL, messageBodyCntr_sram
	lds		wriH, messageBodyCntr_sram+1
	subi16	wriH, wriL, STK500V2_MAX_MSG_SIZE
	brsh	appendByteToMsgBody_end ; Branch if too high.
; messageBodyCntr_sram < STK500V2_MAX_MSG_SIZE.
; Store the byte at current index:
	lds		wriL, messageBodyCntr_sram
	lds		wriH, messageBodyCntr_sram+1
	ldiz	messageBody_sram
	addz16	wriH, wriL
	st		Z, wri0
; Increment the index:
	adiw	wriL, 1
	sts		messageBodyCntr_sram, wriL
	sts		messageBodyCntr_sram+1, wriH
appendByteToMsgBody_end:
	pop		ZH
	pop		ZL
	pop		wriH
	pop		wriL
	ret

; == Routine: Append program-memory bytes to message body ==
; Parameter: Z, pointing to start of program memory bytes to be appended (destroyed);
;	wri0, count of the bytes to be appended (destroyed).
appendFlashBytesToMsgBody:
	push	wri1
	mov		wri1, wri0
appendFlashBytesToMsgBody0:
	tst		wri1
	breq	appendFlashBytesToMsgBody1
	lpm		wri0, Z+
	rcall	appendByteToMsgBody
	dec		wri1
	rjmp	appendFlashBytesToMsgBody0
appendFlashBytesToMsgBody1:
	pop		wri1
	ret
	
; == Routine: Return answer message to host PC ==
returnAnswer:
	push	wri0
	push	wri1
	push	wriL
	push	wriH
	push	ZL
	push	ZH
	clr		wri1 ; Becomes the checksum.
; Send start byte:
	ldi		wri0, MESSAGE_START
	eor		wri1, wri0
	rcall	sendStkByte
; Send sequence number:
	lds		wri0, sequenceNumber_sram
	eor		wri1, wri0
	rcall	sendStkByte
; Send message size:
	lds		wriH, messageBodyCntr_sram+1
	mov		wri0, wriH
	eor		wri1, wri0
	rcall	sendStkByte
	lds		wriL, messageBodyCntr_sram
	mov		wri0, wriL
	eor		wri1, wri0
	rcall	sendStkByte
; Send token:
	ldi		wri0, TOKEN
	eor		wri1, wri0
	rcall	sendStkByte
; Send message body:
	ldiz	messageBody_sram
returnAnswer0:
	ld		wri0, Z+
	eor		wri1, wri0
	rcall	sendStkByte
	sbiw	wriL, 1
	mov		wri0, wriH
	or		wri0, wriL
	brne	returnAnswer0
; Send checksum:
	mov		wri0, wri1
	rcall	sendStkByte
	pop		ZH
	pop		ZL
	pop		wriH
	pop		wriL
	pop		wri1
	pop		wri0
	ret

; == Routine: poll RDY/BSY ==
pollRdyBsy:
	push	wri0
pollRdyBsy0:
	ldi		wri0, 0xF0 ; ISP instruction byte #1
	rcall	spiTxRx
	ldi		wri0, 0x00
	rcall	spiTxRx ; ISP instruction byte #2
	rcall	spiTxRx ; ISP instruction byte #3
	rcall	spiTxRx ; ISP instruction byte #4
	andi	wri0, 0x01 ; Evaluate return value.
	brne	pollRdyBsy0
	pop		wri0
	ret

; == UART RXC (Receive Complete) ISR ==
.equ STK500V2_STATE_CNT = 7 ; The count of all protocol states.
uartRxcIsr:
	push	wr0
	xin		wr0, CPU_SREG
	push	wri0
	push	wri1
	push	wriL
	push	wriH
	push	ZL
	push	ZH
; Get the received byte:
	xin		wri0, USART0_RXDATAL ; Clears the interrupt flag.
; Branch according to stk500v2 protocol state:
	lds		wri1, stk500v2State_sram
	cpi		wri1, STK500V2_STATE_CNT
	brsh	uartRxcIsr0 ; Branch if state is invalid.
uartRxcIsr2:
	ldiz	uartRxcIsr1
	addz	wri1
	ijmp
uartRxcIsr0:
	sts		stk500v2State_sram, zero
	clr		wri1
	rjmp	uartRxcIsr2
uartRxcIsr1:
	rjmp	uartRxcIsr_state0 ; Expecting start byte.
	rjmp	uartRxcIsr_state1 ; Expecting sequence number.
	rjmp	uartRxcIsr_state2 ; Expecting message body size MSB.
	rjmp	uartRxcIsr_state3 ; Expecting message body size LSB.
	rjmp	uartRxcIsr_state4 ; Expecting token.
	rjmp	uartRxcIsr_state5 ; Expecting data (message body) byte.
	rjmp	uartRxcIsr_state6 ; Expecting checksum.
	
uartRxcIsr_state0:
	cpi		wri0, MESSAGE_START
	brne	uartRxcIsr_state01
	sts		checksum_sram, wri0
	ldi		wri0, 1
	sts		stk500v2State_sram, wri0
uartRxcIsr_state01:
	rjmp	uartRxcIsr_end
	
uartRxcIsr_state1:
	lds		wri1, checksum_sram
	eor		wri1, wri0
	sts		checksum_sram, wri1
	sts		sequenceNumber_sram, wri0
	ldi		wri0, 2
	sts		stk500v2State_sram, wri0
	rjmp	uartRxcIsr_end
	
uartRxcIsr_state2:
	lds		wri1, checksum_sram
	eor		wri1, wri0
	sts		checksum_sram, wri1
	sts		messageSize_sram+1, wri0
	ldi		wri0, 3
	sts		stk500v2State_sram, wri0
	rjmp	uartRxcIsr_end
	
uartRxcIsr_state3:
	lds		wri1, checksum_sram
	eor		wri1, wri0
	sts		checksum_sram, wri1
	sts		messageSize_sram, wri0
	ldi		wri0, 4
	sts		stk500v2State_sram, wri0
	rjmp	uartRxcIsr_end
	
uartRxcIsr_state4:
	cpi		wri0, TOKEN
	brne	uartRxcIsr_state41
	lds		wri1, checksum_sram
	eor		wri1, wri0
	sts		checksum_sram, wri1
	sts		messageBodyCntr_sram, zero
	sts		messageBodyCntr_sram+1, zero
	ldi		wri0, 5
	sts		stk500v2State_sram, wri0
	rjmp	uartRxcIsr_end
uartRxcIsr_state41:
	sts		stk500v2State_sram, zero
	rjmp	uartRxcIsr_end
	
uartRxcIsr_state5:
; Check whether messageBodyCntr_sram is in allowed range:
	lds		wriL, messageBodyCntr_sram
	lds		wriH, messageBodyCntr_sram+1
	subi16	wriH, wriL, STK500V2_MAX_MSG_SIZE
	brsh	uartRxcIsr_state51 ; Branch if too high.
; messageBodyCntr_sram < STK500V2_MAX_MSG_SIZE.
	lds		wri1, checksum_sram
	eor		wri1, wri0
	sts		checksum_sram, wri1
; Store the received byte at the current index:
	lds		wriL, messageBodyCntr_sram
	lds		wriH, messageBodyCntr_sram+1
	ldiz	messageBody_sram
	addz16	wriH, wriL
	st		Z, wri0
; Increment the index:
	adiw	wriL, 1
	sts		messageBodyCntr_sram, wriL
	sts		messageBodyCntr_sram+1, wriH
; Test whether this was the last message body byte:
	lds		ZL, messageSize_sram
	lds		ZH, messageSize_sram+1
	sub		wriL, ZL
	sbc		wriH, ZH
	brlo	uartRxcIsr_state52 ; Branch if end not reached yet.
	ldi		wri0, 6
	sts		stk500v2State_sram, wri0
uartRxcIsr_state52:
	rjmp	uartRxcIsr_end
uartRxcIsr_state51:
	sts		stk500v2State_sram, zero
	rjmp	uartRxcIsr_end
	
uartRxcIsr_state6:
	sts		stk500v2State_sram, zero
	set
	lds		wri1, checksum_sram
	cp		wri0, wri1
	brne	uartRxcIsr_state61
	ldi		event, EVENT_RX_SUCCESS
	rjmp	uartRxcIsr_end
uartRxcIsr_state61:
	ldi		event, EVENT_CKSUM_ERROR
	
uartRxcIsr_end:
	pop		ZH
	pop		ZL
	pop		wriH
	pop		wriL
	pop		wri1
	pop		wri0
	xout	CPU_SREG, wr0
	pop		wr0
	reti
	
; == TCA0 Overflow ISR ==
tcaOvfIsr:
	push	wr0
	xin		wr0, CPU_SREG
	push	wri0
; Clear the interrupt flag:
	ldi		wri0, TCA_SINGLE_OVF_bm
	xout	TCA0_SINGLE_INTFLAGS, wri0
; Decrement register 'countdown' if not already zero:
	tst		countdown
	breq	tcaOvfIsr0
	dec		countdown
tcaOvfIsr0:
	pop		wri0
	xout	CPU_SREG, wr0
	pop		wr0
	reti
	